import zipfile, io, json, os
import numpy as np
import random
from threading import Lock
import functools
import datetime
from pathlib import Path
import linecache
import pyarrow as pa
import pyarrow.dataset as ds

THIS_DIR = Path(__file__).absolute().parent

class Session:
    def __init__(self, id, schema=None):
        self.id = id
        self.name = id.split("-")[2]
        self.timestamp = id.split("-")[1]
        self.schema = None
        self.value_dict = dict()

        if schema:
            self.schema = schema
            cache_path = os.path.join(".zcache", f"{self.id}.arrow")
            os.makedirs(".zcache", exist_ok=True)
            self.sink = pa.OSFile(cache_path, 'wb')
            self.writer = pa.ipc.new_file(self.sink, self.schema)

    # object destructor since we are being risky with our file handles
    def __del__(self):
        self.close_session()

    def close_session(self):
        self.writer.close()
        self.sink.close()

    def append(self, **kwargs):
        arrays = list()
        for arg, value in kwargs.items():
            try:
                field = self.schema.field(arg)
                array = pa.array(value, type=field.type)
            except pa.ArrowInvalid as e:
                field_name = field.name
                raise ValueError(f"Failed to convert field {field_name} to {field.type}")
            arrays.append(array)

        batch = pa.record_batch(arrays, self.schema)
        print("Writing ", batch)
        self.writer.write(batch)

class Monitor():
    def __init__(self, connection, compression=zipfile.ZIP_DEFLATED):
        self.connection = connection
        self.memory = None
        self.compression = compression
        self._reset()

        self.session_data = {}
        self._session_locks = {}
        self._active_session_key = None

    def __enter__(self):
        return self

    def close(self):
        self.memory.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _reset(self):
        self.names = set()
        self.header = {"inputs": [], "outputs": []}
        if self.memory is not None:
            self.memory.close()
        self.memory = io.BytesIO(b'')

    def _add_array(self, is_input, name, numpy_array, named_array):
        if not isinstance(numpy_array, np.ndarray):
            raise Exception("Expected a numpy array. Got: " + type(numpy_array))

        instance = {"name": name,
                    "shape": numpy_array.shape,
                    "type": numpy_array.dtype.str,
                    "named": False}

        if named_array is not None:
            if not isinstance(named_array, np.ndarray):
                raise Exception("Expected a numpy array. Got: " + type(named_array))
            if named_array.dtype.str[1] != "U":
                raise Exception("Expected np._object. Got: " + named_array.dtype)
            if len(numpy_array.shape) != named_array.size:
                if len(numpy_array.shape) != len(named_array.shape):
                    raise Exception("Invalid named array")
                for i in range(len(numpy_array.shape)):
                    if numpy_array.shape[i] % named_array.shape[i] != 0:
                        raise Exception("Invalid named array")
            instance["named"] = True

        if name not in self.names:
            if is_input:
                self.header["inputs"].append(instance)
            else:
                self.header["outputs"].append(instance)
            self.names.add(name)
        else:
            raise Exception("The array \'" + name + "\' already exists")


        with zipfile.ZipFile(self.memory, 'a', self.compression) as file:
            file.writestr(name, numpy_array.tobytes())
            if named_array is not None:
                file.writestr(name + "_named", named_array.tobytes())

    def generate_key(self):
        return f"session-{datetime.datetime.now().strftime('%Y%m%d%H%M%S')}-{self.gen_words()}"

    def gen_words(self):
        filename = 'utils/eff-short.txt'
        filepath = THIS_DIR / filename
        line_count = len(linecache.getlines(str(filepath)))
        word_count = 4
        words = list()

        for i in range(word_count):
            rand = random.randint(0, line_count)
            word = linecache.getline(str(filepath), rand)
            words.append(word.rstrip())

        return "-".join(words)

    def get_session_lock(self, key):
        if key not in self._session_locks:
            self._session_locks[key] = Lock()
        return self._session_locks[key]

    def get_current_session_key(self):
        if self._active_session_key is not None:
            return self._active_session_key
        else:
            return self.open_session()

    def open_session(self, schema=None):
        key = self.generate_key()
        self._active_session_key = key
        self.session_data[key] = Session(key, schema)
        return key

    def close_session(self, key):
        print(f"Closing session {key}")
        del self.session_data[key]
        self._active_session_key = None

    def session(self, schema=None):
        def wrap(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                key = self.open_session(schema)
                try:
                    return func(*args, **kwargs)
                finally:
                    self.close_session(key)
            return wrapper
        return wrap

    def log(self, *args, **kwargs):
        key = self.get_current_session_key()
        if key is None:
            raise Exception("Not in session")

        lock = self.get_session_lock(key)
        with lock:
            self.session_data[key].append(*args, **kwargs)

    def add_input(self, name, numpy_array, named_array=None):
        self.connection.auth_check()
        self._add_array(True, name, numpy_array, named_array)

    def add_output(self, name, numpy_array, named_array=None):
        self.connection.auth_check()
        self._add_array(False, name, numpy_array, named_array)

    def send(self, name, org=None, project=None):
        self.connection.auth_check()
        with zipfile.ZipFile(self.memory, 'a', self.compression) as file:
            file.writestr("header.json", json.dumps(self.header))

        self.connection.config(project=project, org=org)

        if name[-4:] != ".zip":
            zip_name = name + ".zip"
        else:
            zip_name = name

        body = {"filename": zip_name, "file_size": self.memory.getbuffer().nbytes, "metadata": json.dumps("")}
        newObj = {
            'upload_status': {'status': "Pending"},
            'dataset_type': 'classification',
        }

        body = {**body, **newObj}
        post_url = self.connection.api_url_builder(self.connection.org, self.connection.project, "tensor")
        res = self.connection.req_handle(post_url, "post", body)

        dataset_id = res.json()["id"]
        res = self.connection.multipart_upload("tensors", dataset_id, self.memory)
        res = self.connection.confirm_upload("tensors", dataset_id)

        print("Completed")
        return res



import threading, os, unittest
import zetane
from .test_base import BaseTestCase, SCRIPT_DIR, ZETANE_DIR
import pyarrow as pa

class TestMonitor(BaseTestCase):
  def tearDown(self) -> None:
    print("teardown..")
    cache_dir = os.path.join(ZETANE_DIR, ".zcache")
    for file in os.listdir(cache_dir):
      print(file)
      if file.endswith(".arrow") or file.endswith(".parquet"):
        file_path = os.path.join(cache_dir, file)
        os.remove(file_path)
        print(f"Deleted {file_path}")
    super().tearDown()

  def test_log_data(self):
    schema = pa.schema([pa.field('text', pa.string())])

    @zetane.session(schema=schema)
    def process():
      zetane.log(text=["Event 1"])
      zetane.log(text=["Event 2"])
      key = list(zetane.monitor_proxy.session_data.keys())[0]
      self.assertEqual(zetane.monitor_proxy.session_data[key], ["Event 1", "Event 2"])

    process()

  def test_multithread(self):
    schema = pa.schema([pa.field('text', pa.string())])
    @zetane.session(schema=schema)
    def process():
      thread = threading.Thread(target=log_thread)
      thread.start()
      zetane.log(text=["Main Thread"])
      thread.join()

      key = list(zetane.monitor_proxy.session_data.keys())[0]
      # note misleadingly named method that used to be called "assertItemsEqual",
      # which is what this actually does
      self.assertCountEqual(zetane.monitor_proxy.session_data[key], ["Main Thread", "Child Thread"])

    def log_thread():
      zetane.monitor_proxy.log(text=["Child Thread"])

    process()


if __name__ == '__main__':
  unittest.main()



import os, filetype, time, json
from tqdm import tqdm
import requests as req
from pathlib import Path
from zetane.utils import validate_test_json, validate_file_type


def get_default_org(json):
    if len(json['organizations']) > 0:
        return json['organizations'][0]['name']

    raise ValueError("No default organization for this user!")

def file_helper(file_path, metadata, name=None):
    filename = os.path.basename(file_path)
    absolute_file_path = os.path.abspath(file_path)
    file_size = os.path.getsize(absolute_file_path)
    file_type = filetype.guess_mime(absolute_file_path)
    if not name:
        name = filename
    body = {"name": name, "filename": filename, "file_type": file_type, "file_size": file_size, "source": "API", "metadata": json.dumps(metadata)}
    newObj = {
            'upload_status': { 'status': "Pending" },
            'dataset_type': 'classification',
        }

    body = {**body, **newObj}
    return body


class NetworkError(Exception):
    def __init__(self, result, message=""):
        self.result = result
        self.message = message
        super().__init__(message)

    def __str__(self):
        return self.result.reason + ' ' + str(self.result.status_code) + ' at ' + self.result.url + ' -> ' + self.result.text

class Connection:
    def __init__(self, api_key, address="https://protector-api-1.zetane.com"):
        self.api_key = api_key
        self.address = address if address.endswith('/') else f'{address}/'
        self.user = None
        self.org = None
        self.project = None

    def confirm_upload(self, client, datatype, id):
        # PUT /org/project/datatype/id
        body = {"upload_status": {"status": "Ready"}}
        confirm_url = self.api_url_builder(client.org(), client.project(), datatype, f"{str(id)}")
        return self.req_handle(confirm_url, "put", body, json_bool=True)

    def multipart_upload(self, datatype, id, file):
        FILE_CHUNK_SIZE = 10000000  # 10MB
        if isinstance(file, str):
            absolute_file_path = os.path.abspath(file)
            file_size = os.path.getsize(absolute_file_path)
            file_path = Path(absolute_file_path)
            file_type = file_path.suffix
            file_obj = open(absolute_file_path, "rb")
        else:
            file_size = file.getbuffer().nbytes
            file_type = "application/zip"
            file_obj = file

        NUM_CHUNKS = (file_size - 1) // FILE_CHUNK_SIZE + 1

        base_url = self.api_url_builder(self.org, self.project, datatype, f"{str(id)}")
        res = self.req_handle(base_url + '/upload', "post", {"fileType": file_type})

        # Initialize multi-part
        upload_id = res.json()["uploadId"]

        # Upload multi-part
        parts = []
        for index in tqdm(range(NUM_CHUNKS)):
            offset = index * FILE_CHUNK_SIZE
            file_obj.seek(offset, 0)

            res = self.req_handle(base_url + "/upload_part",
                            "post",
                        {"part": index + 1, "uploadId": upload_id})
            presigned_url = res.json()["presignedUrl"]
            res = req.put(presigned_url, data=file_obj.read(FILE_CHUNK_SIZE), headers={"Content-Type": file_type})
            parts.append({"ETag": res.headers["etag"][1:-1], "PartNumber": index + 1})

        if isinstance(file, str):
            file_obj.close()
        # Finalize multi-part
        fin_body = {"parts": parts, "uploadId": upload_id}
        res = self.req_handle(base_url + "/upload_complete", "post", fin_body, json_bool=True)
        return res

    def req_handle(self, url, method, body=None, json_bool=False):
        headers = {"Authorization": "Token " + self.api_key}
        res = None
        url = self.address + url
        if method == "get":
            res = req.get(url, headers=headers)
        if method == "post":
            if json_bool:
                res = req.post(url, json=body, headers=headers)
            else:
                res = req.post(url, data=body, headers=headers)
        if method == "put":
            if json_bool:
                res = req.put(url, json=body, headers=headers)
            else:
                res = req.put(url, data=body, headers=headers)
        if method == "delete":
            res = req.delete(url, data=body, headers=headers)

        #if res is not None and res.status_code != 200 and res.status_code != 201 and res.status_code != 202:
        #    print('STATUS CODE: ', res.status_code)
        #    raise NetworkError(res)

        return res

    def api_url_builder(self, *args):
        url = 'api'
        for arg in args:
            url = url + "/" + arg
        return url

    def refresh_user(self):
        try:
            res = self.req_handle("users/me", 'get')
            self.user = res.json()
        except:
            raise Exception("API KEY is invalid. You can retrieve your api key from: protector.zetane.com")

    def auth_check(self):
        if not hasattr(self, 'api_key') or not self.api_key:
            raise SystemExit('Failed to authenticate API key')

    def check_org_in_user(self, org):
        user = self.user
        for org_temp in user['organizations']:
            if org_temp['name'] == org:
                return True
        print('Organization name is invalid')
        return False

    def check_project_in_org(self, org, project_to_check):
        user = self.user
        org_check = None
        for org_tmp in user['organizations']:
            if org_tmp['name'] == org:
                org_check = org_tmp
        for project in org_check['projects']:
            if project['name'] == project_to_check:
                return True
        print('Project name is invalid')
        return False

    def config(self, api_key=None, address=None, org=None, project=None):
        self.refresh_user()

        if org and self.check_org_in_user(org.lower()):
            self.org = org.lower()
            print('Organization configuration successful')
        else:
            self.org = get_default_org(self.user)
        if self.org and project and self.check_project_in_org(self.org, project):
            self.project = project
            print('Project configuration successful')

class Protector():
    def __init__(self, connection):
        self.connection = connection
        print('Successfully authenticated: ')
        self.get_orgs_and_projects()
        self.model_filename = None
        self.dataset_filename = None

    def api_key(self):
        return self.connection.api_key

    def auth_check(self):
        return self.connection.auth_check()

    def config(self, *args, **kwargs):
        return self.connection.config(*args, **kwargs)

    def org(self):
        return self.connection.org

    def project(self):
        return self.connection.project

    def address(self):
        return self.connection.address

    def api_url_builder(self, *args, **kwargs):
        return self.connection.api_url_builder(*args, **kwargs)

    def req_handle(self, *args, **kwargs):
        return self.connection.req_handle(*args, **kwargs)

    def get_orgs_and_projects(self):
        self.auth_check()
        res = self.req_handle('users/me', 'get')
        user = res.json()
        for org in user['organizations']:
            print("Organization: ", org['name'])
            for project in org['projects']:
                print('\t' + "Project: ",  project['name'])

    def create_project(self, name):
        self.auth_check()
        print("Creating project")
        body = {"name": name}
        create_url = self.api_url_builder(self.org(), "project")
        res = self.req_handle(create_url, "post", body)
        if (res.status_code != 201):
            raise RuntimeError("Project was not created : ", res.json())
        return res

    def delete_project(self, name):
        self.auth_check()
        delete_url = self.api_url_builder(self.org(), name, "delete")
        return self.req_handle(delete_url, "delete")

    def upload_dataset(self, file_path='', name=None, org=None, project=None):
        # POST /org/project/dataset
        self.auth_check()
        print('Starting dataset upload..')
        self.config(project=project, org=org)
        # metadata = construct_metadata_obj(file_path)
        # if not metadata:
        #     print('Please select a zipped file to upload')
        #     return
        # return
        body = file_helper(file_path, {}, name)
        post_url = self.api_url_builder(self.org(), self.project(), "dataset")
        res = self.req_handle(post_url, "post", body, True)
        if res.status_code == 201:
            dataset_id = res.json()["id"]
            dataset_filename = res.json()['filename']
            res = self.connection.multipart_upload("datasets", dataset_id, file_path)
            res = self.connection.confirm_upload(self, "datasets", dataset_id)
            self.dataset_filename = dataset_filename
            print("Completed")
        else:
            raise RuntimeError(f"Dataset did not successfully upload. Error: {res.json()}")
        return res

    def upload_model(self, file_path, name=None, org=None, project=None):
        self.auth_check()
        print('Starting model upload...')
        print("Upload to: ", project)
        self.config(project=project, org=org)
        # metadata = construct_metadata_obj(file_path)
        # if not metadata:
        #     print('Please select a zipped file to upload')
        #     return
        body = file_helper(file_path, {}, name)
        print("Project: ", self.project())
        post_url = self.api_url_builder(self.org(), self.project(), "model")
        res = self.req_handle(post_url, "post", body, json_bool=True)
        if res.status_code == 201:
            model_id = res.json()["id"]
            model_filename = res.json()['filename']
            res = self.connection.multipart_upload("models", model_id, file_path)
            res = self.connection.confirm_upload(self, "models", model_id)
            self.model_filename = model_filename
            print("Completed")
        else:
            raise RuntimeError(f"Model did not successfully upload. Error: {res.json()}")
        return res

    def get_entries(self, datatype, org=None, project=None):
        # GET /org/project/datatype
        self.auth_check()
        print("Getting entries..")
        if datatype != 'models' and datatype != 'datasets':
            return print('Only available datatypes are "models" or "datasets"')
        self.config(project=project, org=org)
        if not self.project():
            return print('Please select a project')
        url = self.api_url_builder(self.org(), self.project(), datatype)
        res = self.req_handle(url, "get")

        if res.status_code == 200:
            for x in res.json():
                if not x['name']:
                    continue
                print(f'ID: {str(x["id"])} NAME: {x["name"]}')
        return res

    def get_report_status(self, name, org=None, project=None):
        # GET /org/project/report_name
        self.auth_check()
        self.config(project=project, org=org)
        url = self.api_url_builder(self.org(), self.project(), f"{name}", "status")
        res = self.req_handle(url, "get")
        if res:
            curr_status = res.json()["status"]
            print('Report Name: ' + name)
            print('Report Status: ' + curr_status)
        return res

    def report(self, test_profile_path=None, test_json=dict(), input_shape=list(), model_type=None, organization=None, project=None, model=None, dataset=None, autoRun=False):
        # POST /org/project/run
        self.auth_check()
        self.config(org=organization, project=project)
        model_filename = model if model else self.model_filename
        dataset_filename = dataset if dataset else self.dataset_filename

        validation_errors = list()
        if not input_shape:
           validation_errors.append('Please specify a model input_shape as a list in the format [B H W C]')
        if not model_type and model_type not in ['object_detection', 'image_classification']:
            validation_errors.append('Please specify a model input type: either "object_detection" or "image_classification"')
        if not model and not hasattr(self, 'model'):
            validation_errors.append('Please specify which model you wish to use')
        if not dataset and not hasattr(self, 'dataset'):
            validation_errors.append('Please specify which dataset you wish to use')
        if not project and not hasattr(self, 'project'):
            validation_errors.append('Please specify which project you wish to use')
        if not test_json and validate_file_type(test_profile_path):
            validation_errors.append('Please provide a json test spec or a path to a file containing a json test spec')

        if len(validation_errors) > 0:
            for error in validation_errors:
                print(error)
            return

        #get model ID
        try:
            url = self.api_url_builder(self.org(), self.project(), 'models')
            res = self.req_handle(url, "get")
            if not res.status_code == 200:
                print('Report creation failed - invalid model selection')
                print(res.json())
                return
            else:
                models_json = res.json()
                models_match = [v['id'] for v in models_json if v['filename'] == model_filename]
                model_id = max(models_match)
        except:
            print('Invalid model selection - please select from the following or upload a new model:')
            return self.get_entries('models')

        #get dataset ID
        try:
            url = self.api_url_builder(self.org(), self.project(), 'datasets')
            res = self.req_handle(url, "get")
            if not res.status_code == 200:
                print('Report creation failed - invalid dataset selection')
                print(res.json())
            else:
                dataset_json = res.json()
                datasets_match = [v['id'] for v in dataset_json if v['filename'] == dataset_filename]
                dataset_id = max(datasets_match)
        except:
            print("Invalid dataset selection - please select from the following or upload a new dataset:")
            return self.get_entries('datasets')

        test_series_name = "Test Report"
        if test_profile_path is not None:
            test_series_name = test_profile_path.split('/')[-1]
            with open(test_profile_path) as f:
                test_json = json.load(f)

        crafted_test_profile = validate_test_json(test_json, model_filename, model_id,input_shape, model_type, dataset_filename, dataset_id, test_series_name)

        url = self.api_url_builder(self.org(), self.project(), "run")
        body = {"save": False, "data": crafted_test_profile, "supValues": { "numPrimaryTests": '', "numComposedTests": '', "totalTestRuns": '', "totalXaiApplied": '' }}
        res = self.req_handle(url, "post", body, json_bool=True)
        name = None
        name = res.json()["name"]

        if res.status_code == 201:
            print("Starting report: " + name)
            print('When completed, your report can be found at the following url:')
            link_url = f"https://protector.zetane.com/{self.org()}/{self.project()}/runs/{name}"
            print(link_url)
            if (autoRun):
                i = 0
                sleep_time = 10
                status_url = self.api_url_builder(self.org(), self.project(), name, "status")
                while True:
                    time.sleep(sleep_time)
                    i += 1
                    res = self.req_handle(status_url, "get")
                    if res.status_code == 200:
                        print("Running... " + str(sleep_time * i) + " seconds")
                        status = res.json()["status"]
                        if status == "Ready":
                            print(f"Report completed! View the results {link_url}")
                            return True
                        elif status == "Error":
                            print(f"Error running report! View the error logs: {res.json()['error']}")
                            return False
        else:
            raise NetworkError(res, "Failed to schedule report")

        return res


    def build_image(self, model_id):
        # POST /org/project/id/image
        url = self.api_url_builder(self.org(), self.project(), f"{str(model_id)}", "image")
        res = self.req_handle(url, "post", json.dumps({id: id}))
        if res.status_code == 201:
            name = res.json()["name"]
            i = 0
            while True:
                time.sleep(10)
                i += 1
                url = self.api_url_builder(name, "image", "status")
                res = self.req_handle(url, "get")
                if res.status_code == 200:
                    print("Building... " + str(10 * i) + " seconds")
                    status = res.json()["status"]["status"]
                    if status != "Running" and status != "Pending":
                        break
                else:
                    print('Not building image...')
                    break
        return res






import unittest
from transformers import DetrImageProcessor, DetrForObjectDetection
import torch
from PIL import Image
import requests
import zetane
from .test_base import BaseTestCase

class TestDetr(BaseTestCase):
  @zetane.session
  def test_write_data(self):
    url = "http://images.cocodataset.org/val2017/000000039769.jpg"
    image = Image.open(requests.get(url, stream=True).raw)

    processor = DetrImageProcessor.from_pretrained("facebook/detr-resnet-50")
    model = DetrForObjectDetection.from_pretrained("facebook/detr-resnet-50")

    inputs = processor(images=image, return_tensors="pt")
    outputs = model(**inputs)

    # convert outputs (bounding boxes and class logits) to COCO API
    # let's only keep detections with score > 0.9
    target_sizes = torch.tensor([image.size[::-1]])
    results = processor.post_process_object_detection(outputs, target_sizes=target_sizes, threshold=0.9)[0]

    zetane.log(**inputs, **results)

    for score, label, box in zip(results["scores"], results["labels"], results["boxes"]):
        box = [round(i, 2) for i in box.tolist()]
        print(
                f"Detected {model.config.id2label[label.item()]} with confidence "
                f"{round(score.item(), 3)} at location {box}"
        )



import unittest
import sys, os
from pathlib import Path
from dotenv import load_dotenv, dotenv_values

SCRIPT_DIR = str(Path(__file__).absolute().parent)
ZETANE_DIR = str(Path(__file__).absolute().parent.parent)
sys.path.append(ZETANE_DIR)

import zetane

class BaseTestCase(unittest.TestCase):
  def setUp(self):
    load_dotenv(SCRIPT_DIR + "/.env")
    api_key = os.getenv("ZETANE_API_KEY", None)
    address = os.getenv("ZETANE_ADDRESS", None)
    zetane.config(api_key=api_key, address=address)



from .protector import Protector, Connection
from .monitor import Monitor
import os

""" Settings """
default_client = None
default_connection = None
monitor_proxy = None

# called on import
api_key = os.getenv("ZETANE_API_KEY", None)
address = os.getenv("ZETANE_ADDRESS", "https://protector-api-1.zetane.com/")

def config(*args, **kwargs):
  return _proxy('config', *args, **kwargs)

def upload_dataset(*args, **kwargs):
  return _proxy('upload_dataset', *args, **kwargs)

def upload_model(*args, **kwargs):
  return _proxy('upload_model', *args, **kwargs)

def get_entries(*args, **kwargs):
  return _proxy('get_entries', *args, **kwargs)

def get_orgs_and_projects(*args, **kwargs):
  return _proxy('get_orgs_and_projects', *args, **kwargs)

def report(*args, **kwargs):
  return _proxy('report', *args, **kwargs)

def get_report_status(*args, **kwargs):
  return _proxy('get_report_status', *args, **kwargs)

def create_project(*args, **kwargs):
  return _proxy('create_project', *args, **kwargs)

def delete_project(*args, **kwargs):
  return _proxy('delete_project', *args, **kwargs)

def log(*args, **kwargs):
  return _monitor_proxy('log', *args, **kwargs)

def session(*args, **kwargs):
  return _monitor_proxy('session', *args, **kwargs)

def get_connection(*args, **kwargs):
  global default_connection

  if not default_connection:
    default_connection = _connection_factory(*args, **kwargs)

  return default_connection

def _proxy(method, *args, **kwargs):
  """Create an analytics client if one doesn't exist and send to it."""
  global default_client
  global default_connection

  if not default_connection:
    default_connection = _connection_factory(*args, **kwargs)

  if not default_client:
    default_client = Protector(default_connection)

  fn = getattr(default_client, method)
  return fn(*args, **kwargs)

def _monitor_proxy(method, *args, **kwargs):
  """Create an analytics client if one doesn't exist and send to it."""
  global monitor_proxy
  global default_connection

  if not default_connection:
    default_connection = _connection_factory(*args, **kwargs)

  if not monitor_proxy:
    monitor_proxy = Monitor(default_connection)

  fn = getattr(monitor_proxy, method)
  return fn(*args, **kwargs)

def _connection_factory(*args, **kwargs):
  if api_key is not None:
    conn = Connection(api_key, address=address)

  elif 'api_key' in kwargs:
    conn = Connection(kwargs['api_key'], kwargs.get("address", "https://protector-api-1.zetane.com"))

  elif len(args) > 0:
    conn = Connection(args[0], args[1])

  else:
      print("** Please set your api key with zetane.config(api_key=\"your_api_key\") before calling other methods. **\n")

      raise TypeError('Failed to authenticate API key, no API key provided.')

  return conn

